<?php

namespace Drupal\commerce_sage_payments\EventSubscriber;

use Drupal\commerce_transaction\Entity\TransactionInterface;
use Drupal\state_machine\Event\WorkflowTransitionEvent;

use Drupal\Core\Entity\EntityTypeManagerInterface;

use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Updates the parent payment when a transaction is updated.
 */
class PaymentUpdater implements EventSubscriberInterface {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Constructs a new PaymentUpdater object.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager) {
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events = [
      'commerce_transaction.settle_capture.post_transition' => ['updateForSettled', 0],
      'commerce_transaction.settle_sale.post_transition' => ['updateForSettled', 0],
      'commerce_transaction.settle_credit.post_transition' => ['updateForSettled', 0],
      'commerce_transaction.fail_capture.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.fail_sale.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.fail_credit.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.decline_capture.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.decline_sale.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.decline_credit.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.expire_capture.post_transition' => ['updateForFailed', 0],
      'commerce_transaction.expire_sale_post_transition' => ['updateForFailed', 0],
      'commerce_transaction.expire_credit.post_transition' => ['updateForFailed', 0],
    ];
    return $events;
  }

  /**
   * Updates the parent payment when a transaction is settled.
   *
   * @param \Drupal\state_machine\Event\WorkflowTransitionEvent $event
   *   The transition event.
   */
  public function updateForSettled(WorkflowTransitionEvent $event) {
    $this->updatePayment(
      $event->getEntity(),
      [
        'charge_state' => 'completed',
        'refund_callback' => [$this, 'updateRefundedPayment'],
      ]
    );
  }

  /**
   * Updates the parent payment when a transaction failed to settle.
   *
   * Remote states for failure are Error, Declined and Expired. We do not need
   * separate states for all cases at the payment level; we track this at the
   * transaction level. We simply mark the payment as `capture_failed` or
   * `refund_failed`.
   *
   * Store managers can see that a payment or a refund has failed and they can
   * either view the transaction details in Drupal (UI still to be implemented)
   * or in the Sage Payments Virtual Terminal.
   *
   * @param \Drupal\state_machine\Event\WorkflowTransitionEvent $event
   *   The transition event.
   */
  public function updateForFailed(WorkflowTransitionEvent $event) {
    $this->updatePayment(
      $event->getEntity(),
      [
        'charge_state' => 'capture_failed',
        'refund_state' => 'refund_failed',
      ]
    );
  }

  /**
   * Updates the payment for the given transaction.
   *
   * @param \Drupal\commerce_transaction\Entity\TransactionInterface $transaction
   *   The transaction.
   * @param array $options
   *   An associative array of options. For each transaction type (charge or
   *   refund) a state or a callable must be given. Supported options are:
   *   - charge_state (string, optional): The ID of the new state to set the
   *     payment to when the transaction is a charge.
   *   - charge_callback (callable, optional): The callable to call for
   *     executing custom logic for determining and setting the new payment
   *     state when the transaction is a charge.
   *   - refund_state (string, optional): The ID of the new state to set the
   *     payment to when the transaction is a refund.
   *   - refund_callback (callable, optional): The callable to call for
   *     executing custom logic for determining and setting the new payment
   *     state when the transaction is a refund.
   */
  protected function updatePayment(
    TransactionInterface $transaction,
    array $options
  ) {
    $payment_state = NULL;
    $type = $this->entityTypeManager
      ->getStorage('commerce_transaction_type')
      ->load($transaction->bundle())
      ->getType();

    if (empty($options["{$type}_callback"]) && empty($options["{$type}_state"])) {
      throw new \RuntimeException(sprintf(
        'No state or callback given when updating the payment for transaction with ID "%s" of type "%s".',
        $transaction->id(),
        $type
      ));
    }

    // If we are given a callable we use that; otherwise, we must have been
    // given the new payment state.
    if (!empty($options["{$type}_callback"])) {
      call_user_func($options["{$type}_callback"], $transaction);
      return;
    }
    $payment_state = $options["{$type}_state"];

    // Update the payment entity.
    $payment = $transaction->getPayment();
    $payment->setState($payment_state);
    $this->entityTypeManager
      ->getStorage('commerce_payment')
      ->save($payment);
  }

  /**
   * Updates the payment for settled refunds.
   *
   * @param \Drupal\commerce_transaction\Entity\TransactionInterface $transaction
   *   The transaction.
   */
  protected function updateRefundedPayment(TransactionInterface $transaction) {
    $payment = $transaction->getPayment();
    $old_refunded_amount = $payment->getRefundedAmount();
    $new_refunded_amount = $old_refunded_amount->add($transaction->getAmount());
    // With the current implementation, we block issuing a refund from the UI if
    // there is already a refund pending. Therefore, there shouldn't exist other
    // refunds to account for when calculating the new state for the payment. It
    // should only be determined based on the state and the amount of the
    // current transaction. See
    // `\Drupal\commerce_sage_payments\Plugin\Commerce\PaymentGateway\Sevd::canRefundPayment()`.
    if ($new_refunded_amount->lessThan($payment->getAmount())) {
      $payment->setState('partially_refunded');
    }
    else {
      $payment->setState('refunded');
    }
    $payment->setRefundedAmount($new_refunded_amount);

    $this->entityTypeManager
      ->getStorage('commerce_payment')
      ->save($payment);
  }

}
