<?php

namespace Drupal\commerce_sage_payments\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_order\Entity\OrderInterface;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Exception\OffsiteCancelOnReturnException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OffsitePaymentGatewayBase;
use Drupal\commerce_price\Price;
use Drupal\commerce_price\RounderInterface;
use Drupal\commerce_sage_payments\MachineName\Bundle\Transaction as TransactionBundle;
use Drupal\commerce_transaction\MachineName\Field\Transaction as TransactionField;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;

use KrystalCode\SagePayments\Sdk\DirectApi\ClientBase;
use KrystalCode\SagePayments\Sdk\DirectApi\ClientFactory;
use KrystalCode\SagePayments\Sdk\DirectApi\Resource\Charges as ChargesApi;
use KrystalCode\SagePayments\Sdk\DirectApi\Resource\Credits as CreditsApi;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides the offsite Sage Exchange Virtual Desktop payment gateway.
 *
 * The Sage Exchange Virtual Desktop (legacy Paya Exchange Virtual Desktop)
 * gateway is Sage Payments' offsite payment solution.
 *
 * @CommercePaymentGateway(
 *   id = "sage_payments_sevd",
 *   label = @Translation("Sage Exchange Virtual Desktop"),
 *   display_label = @Translation("Sage Payments (Offsite)"),
 *   modes = {
 *     "test" = @Translation("Sandbox"),
 *     "live" = @Translation("Live"),
 *   },
 *   forms = {
 *     "add-payment-method" = "Drupal\commerce_payment\PluginForm\PaymentMethodAddForm",
 *     "offsite-payment" = "Drupal\commerce_sage_payments\PluginForm\SevdPaymentOffsiteForm",
 *   },
 *   credit_card_types = {
 *     "amex", "discover", "jcb", "mastercard", "visa"
 *   },
 *   payment_method_types = {"credit_card"},
 *   payment_type = "sage_payments",
 *   requires_billing_information = FALSE,
 * )
 */
class Sevd extends OffsitePaymentGatewayBase implements SevdInterface {

  /**
   * The Sage Payments Direct API client factory.
   *
   * @var \KrystalCode\SagePayments\Sdk\DirectApi\ClientFactory
   */
  protected $clientFactory;

  /**
   * The price rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected $rounder;

  /**
   * Constructs a new Sevd object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \KrystalCode\SagePayments\Sdk\DirectApi\ClientFactory $client_factory
   *   The Sage Payments Direct API client factory.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The rounder.
   */
  public function __construct(
    array $configuration,
    $plugin_id,
    $plugin_definition,
    EntityTypeManagerInterface $entity_type_manager,
    PaymentTypeManager $payment_type_manager,
    PaymentMethodTypeManager $payment_method_type_manager,
    TimeInterface $time,
    ClientFactory $client_factory,
    RounderInterface $rounder
  ) {
    parent::__construct(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $entity_type_manager,
      $payment_type_manager,
      $payment_method_type_manager,
      $time
    );

    $this->clientFactory = $client_factory;
    $this->rounder = $rounder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(
    ContainerInterface $container,
    array $configuration,
    $plugin_id,
    $plugin_definition
  ) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('commerce_sage_payments.api.factory'),
      $container->get('commerce_price.rounder')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'credentials' => [
        'application_id' => '',
        'client_id' => '',
        'client_secret' => '',
        'merchant_id' => '',
        'merchant_key' => '',
      ],
      'ui' => [
        'edit_billing' => 0,
      ],
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['credentials'] = $this->buildCredentialsConfigurationForm();
    $form['ui'] = $this->buildUiConfigurationForm();

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    parent::validateConfigurationForm($form, $form_state);

    $values = $form_state->getValue($form['#parents']);

    // If we do not collect billing information in Drupal, we have to allow
    // editing the customer details on the UI.
    if ($values['collect_billing_information']) {
      return;
    }
    if ($values['ui']['edit_billing']) {
      return;
    }

    $form_state->setError(
        $form['ui']['edit_billing'],
        $this->t(
          'You must allow editing billing information on the offiste UI if
           collecting billing within the application is disabled.'
        )
      );
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(
    array &$form,
    FormStateInterface $form_state
  ) {
    parent::submitConfigurationForm($form, $form_state);

    $values = $form_state->getValue($form['#parents']);
    $this->configuration['credentials'] = $values['credentials'];
    $this->configuration['ui'] = $values['ui'];

    // Convert boolean values.
    $ui = &$values['ui'];
    $this->configuration['ui']['edit_billing'] = (bool) $ui['edit_billing'];
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(
    PaymentInterface $payment,
    Price $amount = NULL
  ) {
    $this->assertPaymentState($payment, ['authorization']);

    $amount = $amount ?: $payment->getAmount();
    $amount = $this->rounder->round($amount);
    $charge = [
      'Amounts' => [
        'Total' => $amount->getNumber(),
      ],
    ];

    $charges_api = $this->clientFactory->get(
      ChargesApi::ID,
      $this->buildApiClientConfig()
    );

    // @I Translate Guzzle exceptions to Commerce Payment exceptions
    //    type     : improvement
    //    priority : low
    //    labels   : api, error-handling
    $charges_api->putCharges($payment->getRemoteId(), $charge);

    // The API simply returns a 200 response without details for successful
    // requests. If there has been an error the client would throw an
    // exception. If we therefore arrive here without errors the request must
    // have been successful and we consider the payment captured.
    $payment->set('amount', $amount);
    $payment->setState('capture_pending');
    $payment->save();

    // Update the transaction as well.
    $transaction_storage = $this->entityTypeManager
      ->getStorage('commerce_transaction');
    $transaction = $transaction_storage->loadPrincipalForPayment(
      $payment,
      TransactionBundle::CHARGE
    );
    $transaction_amount = $transaction
      ->get(TransactionField::AMOUNT)
      ->first()
      ->toPrice();
    if (!$amount->equals($transaction_amount)) {
      $transaction->set(TransactionField::AMOUNT, $amount);
    }
    $transaction
      ->get(TransactionField::REMOTE_STATE)
      ->first()
      ->applyTransitionById('capture_authorization');
    $transaction_storage->save($transaction);
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment) {
    $this->assertPaymentState($payment, ['authorization']);

    $charges_api = $this->clientFactory->get(
      ChargesApi::ID,
      $this->buildApiClientConfig()
    );

    $charges_api->deleteCharges($payment->getRemoteId());

    // The API simply returns a 200 response without details for successful
    // requests. If there has been an error the client would throw an
    // exception. If we therefore arrive here without errors the request must
    // have been successful and we consider the payment voided.
    $payment->setState('authorization_voided');
    $payment->save();

    // Void the transaction as well so that it is in the correct state.
    $transaction_storage = $this->entityTypeManager
      ->getStorage('commerce_transaction');
    $transaction = $transaction_storage->loadPrincipalForPayment(
      $payment,
      TransactionBundle::CHARGE
    );
    $transaction
      ->get(TransactionField::REMOTE_STATE)
      ->first()
      ->applyTransitionById('void_authorization');
    $transaction_storage->save($transaction);
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(
    PaymentInterface $payment,
    Price $amount = NULL
  ) {
    $this->assertPaymentState(
      $payment,
      ['completed', 'partially_refunded', 'refund_failed']
    );

    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    // Round the amount as it can be manually entered via the Refund form.
    $amount = $this->rounder->round($amount);
    $this->assertRefundAmount($payment, $amount);

    $credits_api = $this->clientFactory->get(
      CreditsApi::ID,
      $this->buildApiClientConfig()
    );
    $response = $credits_api->postCreditsReference(
      $payment->getRemoteId(),
      ['amount' => $amount->getNumber()]
    );

    // The API simply returns a 200 response without details for successful
    // requests. If there has been an error the client would throw an
    // exception. If we therefore arrive here without errors the request must
    // have been successful and we consider the refund successful.
    // Update the payment.
    $payment->setState('refund_pending');
    $payment->save();

    // Create a transaction for the refund since it is a separate transaction in
    // Sage Payments.
    $transaction_storage = $this->entityTypeManager
      ->getStorage('commerce_transaction');
    $transaction = $transaction_storage->createForPayment(
      $payment,
      [
        TransactionField::TYPE => TransactionBundle::REFUND,
        TransactionField::AMOUNT => $amount,
        TransactionField::REMOTE_ID => $response->reference,
        TransactionField::REMOTE_STATE => 'credit_batch',
      ]
    );
    $transaction_storage->save($transaction);
  }

  /**
   * {@inheritdoc}
   */
  public function canVoidPayment(PaymentInterface $payment) {
    // Return FALSE if the payment is not pending.
    // We do not check for days passed since authorization because Sage Payments
    // allows to delete the authorized transaction regardless. For captures, it
    // will be declined if the days have passed.
    // We could be checking via an API call if the transaction has reached the
    // Authorization Expired status to respect the merchant settings on the
    // portal.
    // @I Prevent voiding or capturing expired authorizations
    //    type     : improvement
    //    priority : low
    //    labels   : sevd
    if ($payment->getState()->getId() !== 'authorization') {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function canCapturePayment(PaymentInterface $payment) {
    return $this->canVoidPayment($payment);
  }

  /**
   * {@inheritdoc}
   *
   * For now, we do not allow issuing a refund if there is already one
   * pending to reduce code complexity. We may improve on that in the future.
   * We'd need to be checking for more pending refunds when processing them and
   * updating the payment to partially or completed refunded. We'd also need to
   * provide a UI for listing the pending transactions associated with a
   * payment.
   */
  public function canRefundPayment(PaymentInterface $payment) {
    return in_array(
      $payment->getState()->getId(),
      ['completed', 'partially_refunded', 'refund_failed'],
      TRUE
    );
  }

  /**
   * {@inheritdoc}
   */
  public function onReturn(OrderInterface $order, Request $request) {
    // Decode the response.
    // @I Move methods for decoding XML responses to the SEVD client
    //    type     : improvement
    //    priority : low
    //    labels   : coding-standards
    $response = $this->openEnvelope($request->request->get('response'));
    $response = $this->parseXmlResponse($response);

    // SEVD does not support a different return URL for when the user cancels
    // the payment. We therefore land here in that case as well. The payload is
    // empty though, so we use that to detect a cancelled payment.
    if (!$response) {
      throw new OffsiteCancelOnReturnException('Cancelled by user.');
    }

    $response = $response['PaymentResponses']['PaymentResponseType'];
    if ($response['Response']['ResponseIndicator'] !== 'A') {
      throw new PaymentGatewayException($response['Response']['Message']);
    }

    $vault = $response['VaultResponse'];
    $transaction_data = $response['TransactionResponse'];

    // Basic validation.
    if ($transaction_data['Reference1'] != $order->id()) {
      throw new PaymentGatewayException(sprintf(
        'The ID of the referenced order (%s) does not match that of the order derived from the route (%s).',
        $transaction_data['Reference1'],
        $order->id()
      ));
    }
    if ($transaction_data['TransactionPaymentType'] !== 'CREDITCARD') {
      throw new PaymentGatewayException(sprintf(
        '%s transaction payment type expected in the response, got %s.',
        'CREDITCARD',
        $transaction_data['TransactionPaymentType']
      ));
    }

    $order_total = $order->getTotalPrice();
    $transaction_data_total = Price::fromArray([
      'number' => $transaction_data['Amount'],
      'currency_code' => $order_total->getCurrencyCode(),
    ]);
    if (!$transaction_data_total->equals($order->getTotalPrice())) {
      throw new PaymentGatewayException(
        'The transaction total does not match the order total.'
      );
    }

    $capture = $order->getData('commerce_sage_payments');
    if (!isset($capture['capture'])) {
      throw new PaymentGatewayException(
        'No information about whether the transaction was a sale or an authorization.'
      );
    }

    // We don't yet support reusing the same payment method; we therefore create
    // a new one for each payment.
    $payment_method_storage = $this->entityTypeManager
      ->getStorage('commerce_payment_method');
    $payment_method = $payment_method_storage->create([
      'type' => 'credit_card',
      'uid' => $order->getCustomerId(),
      'payment_gateway' => $order->get('payment_gateway')->target_id,
      'payment_gateway_mode' => $this->getMode(),
      'remote_id' => $vault['GUID'],
      'card_exp_month' => substr($vault['ExpirationDate'], 0, 2),
      // Don't forget in about 70 years to correct the century.
      'card_exp_year' => '20' . substr($vault['ExpirationDate'], 2, 2),
      'card_number' => substr($vault['Last4'], -4),
      'card_type' => $this->cardTypeMapping()[$vault['PaymentTypeID']],
    ]);
    $billing_profile = $order->getBillingProfile();
    if ($billing_profile) {
      $payment_method->set('billing_profile', $billing_profile);
    }
    $payment_method_storage->save($payment_method);

    // Create the payment.
    $payment_storage = $this->entityTypeManager->getStorage('commerce_payment');
    $payment = $payment_storage->create([
      'type' => 'sage_payments',
      'payment_gateway' => $this->parentEntity->id(),
      'payment_gateway_mode' => $this->getMode(),
      'payment_method' => $payment_method->id(),
      'order_id' => $order->id(),
      'remote_id' => $transaction_data['VANReference'],
      'state' => $capture['capture'] ? 'capture_pending' : 'authorization',
      'amount' => $transaction_data_total,
    ]);
    $payment_storage->save($payment);

    // Create the transaction.
    $transaction_storage = $this->entityTypeManager
      ->getStorage('commerce_transaction');
    $transaction_values = [
      TransactionField::TYPE => TransactionBundle::CHARGE,
      TransactionField::AMOUNT => $transaction_data_total,
      TransactionField::REMOTE_ID => $transaction_data['VANReference'],
      TransactionField::REMOTE_STATE => 'authorization_batch',
    ];
    if ($capture['capture']) {
      $transaction_values[TransactionField::REMOTE_STATE] = 'sale_batch';
    }
    $transaction = $transaction_storage->createForPayment(
      $payment,
      $transaction_values
    );
    $transaction_storage->save($transaction);
  }

  /**
   * Builds the credentials fieldset for the configuration form.
   *
   * @return array
   *   The form element for the credentials fieldset.
   */
  protected function buildCredentialsConfigurationForm(): array {
    $element = [
      '#tree' => TRUE,
      '#type' => 'fieldset',
      '#title' => $this->t('Credentials'),
      '#weight' => 10,
    ];
    $element['application_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Application ID'),
      '#default_value' => $this->configuration['credentials']['application_id'],
      '#maxlength' => 255,
      '#required' => TRUE,
    ];
    $element['client_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Client ID'),
      '#default_value' => $this->configuration['credentials']['client_id'],
      '#maxlength' => 255,
      '#required' => TRUE,
    ];
    $element['client_secret'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Client secret'),
      '#default_value' => $this->configuration['credentials']['client_secret'],
      '#maxlength' => 255,
      '#required' => TRUE,
    ];
    $element['merchant_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant ID'),
      '#default_value' => $this->configuration['credentials']['merchant_id'],
      '#maxlength' => 255,
      '#required' => TRUE,
    ];
    $element['merchant_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Merchant key'),
      '#default_value' => $this->configuration['credentials']['merchant_key'],
      '#maxlength' => 255,
      '#required' => TRUE,
    ];

    return $element;
  }

  /**
   * Builds the UI fieldset for the configuration form.
   *
   * @return array
   *   The form element for the UI fieldset.
   */
  protected function buildUiConfigurationForm(): array {
    $element = [
      '#tree' => TRUE,
      '#type' => 'fieldset',
      '#title' => $this->t('UI capabilities'),
      '#weight' => 11,
    ];
    $element['edit_billing'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Edit billing profile'),
      '#description' => $this->t(
        'When enabled, the customer will be able to change the billing details
         before submitting the payment to Sage Payments. This could result in a
         different billing address to be registered with the order on the
         website than the one registered with Sage Payments.'
      ),
      '#default_value' => $this->configuration['ui']['edit_billing'],
    ];

    return $element;
  }

  /**
   * Returns the configuration for the API client.
   *
   * @return array
   *   An associative array containing the client configuration.
   */
  protected function buildApiClientConfig(): array {
    $config = [
      'client_id' => $this->configuration['credentials']['client_id'],
      'client_secret' => $this->configuration['credentials']['client_secret'],
      'merchant_id' => $this->configuration['credentials']['merchant_id'],
      'merchant_key' => $this->configuration['credentials']['merchant_key'],
    ];

    if ($this->getMode() === 'live') {
      $config['env'] = ClientBase::ENV_PRODUCTION;
    }

    return $config;
  }

  /**
   * Returns the card type mapping between Commerce and SEVD.
   *
   * @return array
   *   An associative array keyed by the SEVD card type ID and having the Drupal
   *   Commerce card type ID as values.
   */
  protected function cardTypeMapping(): array {
    return [
      '3' => 'amex',
      '4' => 'visa',
      '5' => 'mastercard',
      '6' => 'discover',
      '7' => 'jcb',
    ];
  }

  /**
   * Decodes a tokenized SEVD envelope.
   *
   * @param string $xml_response
   *   The SEVD envelope.
   *
   * @return string
   *   The decoded data as an XML string.
   */
  protected function openEnvelope(string $xml_response): string {
    return $this->makePostRequest(
      'request=' . urlencode($xml_response),
      'https://www.sageexchange.com/sevd/frmOpenEnvelope.aspx'
    );
  }

  /**
   * Makes a POST request for decoding data for a tokenized SEVD envelope.
   *
   * @param string $body
   *   The request body.
   * @param string $url
   *   The URL for the request.
   *
   * @return string
   *   The decoded data as an XML string.
   */
  protected function makePostRequest(string $body, string $url): string {
    $context = stream_context_create([
      'http' => [
        'header' => [
          'content-type: application/x-www-form-urlencoded',
          'accept: application/xml',
        ],
        'method' => 'POST',
        'content' => $body,
      ],
    ]);

    return file_get_contents($url, FALSE, $context);
  }

  /**
   * Converts an XML string to a PHP array.
   *
   * @param string $response
   *   The XML response string.
   *
   * @return array|bool
   *   The PHP array, or FALSE if the response was empty.
   */
  protected function parseXmlResponse(string $response) {
    return json_decode(
      json_encode(simplexml_load_string($response)),
      TRUE
    );
  }

}
