<?php

namespace Drupal\commerce_sage_payments\PluginForm;

use Drupal\commerce_payment\PluginForm\PaymentOffsiteForm;
use Drupal\commerce_price\RounderInterface;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\Url;

use KrystalCode\SagePayments\Sdk\Sevd\Request\Charge as ChargeRequest;
use KrystalCode\SagePayments\Sdk\Sevd\ClientFactory;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * The payment process form the offsite gateway.
 *
 * It builds the form that is automatically submitted to SEVD so that the user
 * is redirected to the offsite UI.
 */
class SevdPaymentOffsiteForm extends PaymentOffsiteForm implements
  ContainerInjectionInterface {

  /**
   * The current user account.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $account;

  /**
   * The SEVD client factory.
   *
   * @var \KrystalCode\SagePayments\Sdk\Sevd\ClientFactory
   */
  protected $clientFactory;

  /**
   * The rounder.
   *
   * @var \Drupal\commerce_price\RounderInterface
   */
  protected $rounder;

  /**
   * Constructs a new SevdPaymentOffsiteForm object.
   *
   * @param \Drupal\Core\Session\AccountProxyInterface $account_proxy
   *   The current user account proxy.
   * @param \KrystalCode\SagePayments\Sdk\Sevd\ClientFactory $client_factory
   *   The SEVD client factory.
   * @param \Drupal\commerce_price\RounderInterface $rounder
   *   The rounder.
   */
  public function __construct(
    AccountProxyInterface $account_proxy,
    ClientFactory $client_factory,
    RounderInterface $rounder
  ) {
    // To initialize the Charge client we need the gateway plugin configuration
    // which is not available yet here. We therefore store the client factory
    // and we build the client before using it.
    $this->clientFactory = $client_factory;

    $this->account = $account_proxy->getAccount();
    $this->rounder = $rounder;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('current_user'),
      $container->get('commerce_sage_payments.sevd.factory'),
      $container->get('commerce_price.rounder')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(
    array $form,
    FormStateInterface $form_state
  ) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['#return_url'] = Url::fromRoute(
      'commerce_sage_payments.sevd_return',
      [
        'commerce_order' => $this->entity->getOrder()->id(),
        'step' => 'payment',
      ],
      [
        'absolute' => TRUE,
        // Won't work if the URL is not HTTPS, even on development.
        'https' => TRUE,
      ]
    )->toString();

    $sage_redirect_url = 'https://www.sageexchange.com/sevd/frmPayment.aspx';
    $form['#action'] = $sage_redirect_url;

    $data = [
      'request' => $this->getTokenizedRequest((bool) $form['#capture']),
      'redirect_url' => $form['#return_url'],
      'consumer_initiated' => 'true',
    ];

    $form = $this->buildRedirectForm(
      $form,
      $form_state,
      $sage_redirect_url,
      $data,
      self::REDIRECT_POST
    );

    // When we return to the website, the information posted by SEVD does not
    // include whether the payment is a capture or just an authorization. We
    // therefore store that on the order so that we set the payment status
    // accordingly.
    $order = $this->entity->getOrder();
    $order->setData(
      'commerce_sage_payments',
      ['capture' => $form['#capture']]
    );
    $order->save();

    return $form;
  }

  /**
   * Returns the a tokenized request for a UI-enabled sale.
   *
   * @param bool $capture
   *   `false` to request an Auth charge (authorization only), `true` for a
   *   Sale charge (capture).
   *
   * @return string
   *   The tokenized request string.
   *
   * @throws \InvalidArgumentException
   *   When required items are missing from the transaction details.
   * @throws \GuzzleHttp\Exception\GuzzleException
   *   If the request was unsuccessful.
   *
   * @I Support merchant transaction ID
   */
  protected function getTokenizedRequest(bool $capture): string {
    $plugin_configuration = $this->plugin->getConfiguration();
    $ui_settings = [
      'edit_customer' => $plugin_configuration['ui']['edit_billing'],
    ];

    return $this->buildClient()->uiSale(
      $this->buildTransactionData(),
      $this->buildCustomerData(),
      $capture,
      $ui_settings
    );
  }

  /**
   * Returns the transaction data required for buildling the XML request.
   *
   * @return array
   *   An associative array containing the transaction data.
   *
   * @see KrystalCode\SagePayments\Sdk\Sevd\Request\Charge::uiSale()
   */
  protected function buildTransactionData(): array {
    $order = $this->entity->getOrder();

    return [
      'Reference1' => $order->id(),
      'Amount' => $this->rounder->round($order->getTotalPrice())->getNumber(),
    ];
  }

  /**
   * Returns the customer data required for buildling the XML request.
   *
   * @return array
   *   An associative array containing the customer data.
   *
   * @see KrystalCode\SagePayments\Sdk\Sevd\Request\Charge::uiSale()
   */
  protected function buildCustomerData(): array {
    $billing_address = [
      'country_code' => '',
      'administrative_area' => '',
      'locality' => '',
      'postal_code' => '',
      'address_line1' => '',
      'address_line2' => '',
      'given_name' => '',
      'additional_name' => '',
      'family_name' => '',
    ];
    $billing_profile = $this->entity->getOrder()->getBillingProfile();
    if ($billing_profile && !$billing_profile->get('address')->isEmpty()) {
      $billing_address = $billing_profile->get('address')->first()->getValue();
    }

    return [
      'Name' => [
        'FirstName' => $billing_address['given_name'] ?? '',
        'MI' => $billing_address['additional_name'] ?? '',
        'LastName' => $billing_address['family_name'] ?? '',
      ],
      'Address' => [
        'AddressLine1' => $billing_address['address_line1'],
        'AddressLine2' => $billing_address['address_line2'],
        'City' => $billing_address['locality'],
        'State' => $billing_address['administrative_area'],
        'ZipCode' => $billing_address['postal_code'],
        'Country' => $billing_address['country_code'],
      ],
    ];
  }

  /**
   * Returns the Charge SEVD request client based on gateway configuration.
   *
   * @return \KrystalCode\SagePayments\Sdk\Sevd\Request\Charge
   *   The request client.
   */
  protected function buildClient(): ChargeRequest {
    $credentials = $this->plugin->getConfiguration()['credentials'];

    $config = [
      'application_id' => $credentials['application_id'],
      'client_id' => $credentials['client_id'],
      'client_secret' => $credentials['client_secret'],
      'merchant_id' => $credentials['merchant_id'],
      'merchant_key' => $credentials['merchant_key'],
      'language_id' => $this->account->getPreferredLangcode(),
    ];

    return $this->clientFactory->get(ChargeRequest::ID, $config);
  }

}
